/*
 * Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and
 * international Copyright laws.  Users and possessors of this source code
 * are hereby granted a nonexclusive, royalty-free license to use this code
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOURCE CODE.
 *
 * U.S. Government End Users.   This source code is a "commercial item" as
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of
 * "commercial computer  software"  and "commercial computer software
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995)
 * and is provided to the U.S. Government only as a commercial end item.
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
 * source code with only those rights set forth herein.
 *
 * Any use of this source code in individual and commercial software must
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */

/* Matrix multiplication: P = M * N.
 * Device code.
 */

#ifndef _MATRIXMUL_KERNEL_H_
#define _MATRIXMUL_KERNEL_H_

#include <stdio.h>
#include "matrixmul.h"


////////////////////////////////////////////////////////////////////////////////
//! Matrix multiplication on the device: P = M * N
//! Mw is M's width and Nw is N's width
////////////////////////////////////////////////////////////////////////////////
    __global__ void
matrixMul( float* P, float* M, float* N, int Mw, int Nw)
{
#if 1
    int bx = blockIdx.x;     int by = blockIdx.y;
    int tx = threadIdx.x;    int ty = threadIdx.y;
    __shared__ float Ms[BLOCK_SIZE][BLOCK_SIZE];
    __shared__ float Ns[BLOCK_SIZE][BLOCK_SIZE];

    // ===================================================================
    // Code segment 1
    // Determine the update values for the tile indices in the loop
    // ===================================================================

    int mBegin = Mw * BLOCK_SIZE * by;
    int mEnd   = mBegin + Mw - 1;
    int mStep  = BLOCK_SIZE;
    int nBegin = BLOCK_SIZE * bx;
    int nStep  = BLOCK_SIZE*Nw;
    float Psub = 0.0f;

    // ===================================================================
    // Code segment 2
    // Do matrix-matrix multiplication inside a tile
    // ===================================================================

    for (int m = mBegin, n = nBegin; m <= mEnd; m += mStep, n += nStep) {

        #if 0 //emu==1
        printf ("(m %d, n %d (%d)), Ms(%d, %d): M[%d]=%0.0f, Ns(%d, %d): N[%d]=%0.0f\n",
            m, n, n/Nw, ty, tx, (m + ty*Mw + tx), M[m + ty*Mw + tx],
            ty, tx, (n + ty*Nw + tx), N[n + ty*Nw + tx]);
        #endif

        // Load a tile from M and N into the shared memory arrays
        Ms[ty][tx] = M[m + ty*Mw + tx];
        Ns[ty][tx] = N[n + ty*Nw + tx];

        // Synchronize the threads
        __syncthreads();

        // Multiply the two tiles together, each thread accumulating
        // the partial sum of a single dot product.
        for (int i = 0; i < BLOCK_SIZE; i++) {
            Psub += Ms[ty][i] * Ns[i][tx];
            #if 0 //emu==1
            printf ("    %10.0f += Ms[%d][%d](%5.0f) * Ns[%d][%d](%5.0f)\n",
                Psub, ty, i, Ms[ty][i], i, tx, Ns[i][tx]);
            #endif
        }

        // Synchronize again.
        __syncthreads();
    }

    // ===================================================================
    // Code segment 3
    // Store the data back to global memory
    // ===================================================================

    int p = Nw * BLOCK_SIZE * by + BLOCK_SIZE * bx;
    P[p + Nw * ty + tx] = Psub;

    #if 0 //emu==1
    printf ("P[p + (Nw * ty + tx) = %d + %d] = %f\n", p, Nw * ty + tx, Psub);
    #endif

#else // contains more computation operations inside the loop.
    int width = Mw;
    __shared__ float Ms[BLOCK_SIZE][BLOCK_SIZE];
    __shared__ float Ns[BLOCK_SIZE][BLOCK_SIZE];

    int bx = blockIdx.x;  int by = blockIdx.y;
    int tx = threadIdx.x; int ty = threadIdx.y;

    // Identify the row and column of the P element to work on
    int Row = by * BLOCK_SIZE + ty;
    int Col = bx * BLOCK_SIZE + tx;

    float Pvalue = 0;
    // Loop over the M and N tiles required to compute the P element
    for (int m = 0; m < width/BLOCK_SIZE; ++m) {
        // Coolaborative loading of M and N tiles into shared memory
        Ms[tx][ty] = M[(m*BLOCK_SIZE + tx)*width+Row];
        Ns[tx][ty] = N[Col*width+(m*BLOCK_SIZE + ty)];
        __syncthreads();

        for (int k = 0; k < BLOCK_SIZE; ++k)
            Pvalue += Ms[k][ty] * Ns[tx][k];

        __syncthreads();
    }

    P[Row*width+Col] = Pvalue;

#endif
}

#endif // #ifndef _MATRIXMUL_KERNEL_H_
